import os
import shutil
import tempfile


def _copy_file_or_tree(src, dst, dst_dir=None):
    """
    :return: The path to the copied artifacts, relative to `dst`
    """
    dst_subpath = os.path.basename(os.path.abspath(src))
    if dst_dir is not None:
        dst_subpath = os.path.join(dst_dir, dst_subpath)
    dst_path = os.path.join(dst, dst_subpath)
    if os.path.isfile(src):
        dst_dirpath = os.path.dirname(dst_path)
        if not os.path.exists(dst_dirpath):
            os.makedirs(dst_dirpath)
        shutil.copy(src=src, dst=dst_path)
    else:
        shutil.copytree(src=src, dst=dst_path)
    return dst_subpath


class TempDir(object):
    def __init__(self, chdr=False, remove_on_exit=True):
        self._dir = None
        self._path = None
        self._chdr = chdr
        self._remove = remove_on_exit

    def __enter__(self):
        self._path = os.path.abspath(tempfile.mkdtemp())
        assert os.path.exists(self._path)
        if self._chdr:
            self._dir = os.path.abspath(os.getcwd())
            os.chdir(self._path)
        return self

    def __exit__(self, tp, val, traceback):
        if self._chdr and self._dir:
            os.chdir(self._dir)
            self._dir = None
        if self._remove and os.path.exists(self._path):
            shutil.rmtree(self._path)

        assert not self._remove or not os.path.exists(self._path)
        assert os.path.exists(os.getcwd())

    def path(self, *path):
        return (
            os.path.join("./", *path) if self._chdr else os.path.join(self._path, *path)
        )



from .preprocessing import (
    create_and_save_preprocessing,
    load_serialized_preprocessing,
    Preprocessing,
)

__all__ = [
    create_and_save_preprocessing,
    load_serialized_preprocessing,
    Preprocessing,
]



from abc import ABCMeta, abstractmethod
import os
import posixpath
import shutil
import urllib

import cloudpickle
from loguru import logger
import yaml

from clearbox_wrapper.exceptions import ClearboxWrapperException
from clearbox_wrapper.model import MLMODEL_FILE_NAME, Model
from clearbox_wrapper.pyfunc import add_pyfunc_flavor_to_model, FLAVOR_NAME
from clearbox_wrapper.utils import (
    _copy_file_or_tree,
    _get_default_conda_env,
    _get_flavor_configuration,
    TempDir,
)

CONFIG_KEY_ARTIFACTS = "artifacts"
CONFIG_KEY_ARTIFACT_RELATIVE_PATH = "path"
CONFIG_KEY_ARTIFACT_URI = "uri"
CONFIG_KEY_PYTHON_MODEL = "python_model"
CONFIG_KEY_CLOUDPICKLE_VERSION = "cloudpickle_version"

POSTGRES = "postgresql"
MYSQL = "mysql"
SQLITE = "sqlite"
MSSQL = "mssql"

DATABASE_ENGINES = [POSTGRES, MYSQL, SQLITE, MSSQL]

_INVALID_DB_URI_MSG = (
    "Please refer to https://mlflow.org/docs/latest/tracking.html#storage for "
    "format specifications."
)

_UNSUPPORTED_DB_TYPE_MSG = "Supported database engines are {%s}" % ", ".join(
    DATABASE_ENGINES
)

new_model = Model()


def get_default_conda_env():
    """
    :return: The default Conda environment for MLflow Models produced by calls to
             :func:`save_model() <mlflow.pyfunc.save_model>`
             and :func:`log_model() <mlflow.pyfunc.log_model>` when a user-defined subclass of
             :class:`PythonModel` is provided.
    """
    return _get_default_conda_env(
        additional_conda_deps=None,
        additional_pip_deps=["cloudpickle=={}".format(cloudpickle.__version__)],
        additional_conda_channels=None,
    )


class PythonModel(object, metaclass=ABCMeta):
    """
    Represents a generic Python model that evaluates inputs and produces API-compatible outputs.
    By subclassing :class:`~PythonModel`, users can create customized MLflow models with the
    "python_function" ("pyfunc") flavor, leveraging custom inference logic and artifact
    dependencies.
    """

    def load_context(self, context):
        """
        Loads artifacts from the specified :class:`~PythonModelContext` that can be used by
        :func:`~PythonModel.predict` when evaluating inputs. When loading an MLflow model with
        :func:`~load_pyfunc`, this method is called as soon as the :class:`~PythonModel` is
        constructed.
        The same :class:`~PythonModelContext` will also be available during calls to
        :func:`~PythonModel.predict`, but it may be more efficient to override this method
        and load artifacts from the context at model load time.
        :param context: A :class:`~PythonModelContext` instance containing artifacts that the
                        model can use to perform inference.
        """

    @abstractmethod
    def predict(self, context, model_input):
        """
        Evaluates a pyfunc-compatible input and produces a pyfunc-compatible output.
        For more information about the pyfunc input/output API, see `pyfunc-inference-api`.
        :param context: A :class:`~PythonModelContext` instance containing artifacts that the
                        model can use to perform inference.
        :param model_input: A pyfunc-compatible input for the model to evaluate.
        """


class PythonModelContext(object):
    """
    A collection of artifacts that a :class:`~PythonModel` can use when performing inference.
    :class:`~PythonModelContext` objects are created *implicitly* by the
    :func:`save_model() <mlflow.pyfunc.save_model>` and
    :func:`log_model() <mlflow.pyfunc.log_model>` persistence methods, using the contents
    specified by the ``artifacts`` parameter of these methods.
    """

    def __init__(self, artifacts):
        """
        :param artifacts: A dictionary of ``<name, artifact_path>`` entries, where
        ``artifact_path`` is an absolute filesystem path to a given artifact.
        """
        self._artifacts = artifacts

    @property
    def artifacts(self):
        """
        A dictionary containing ``<name, artifact_path>`` entries, where ``artifact_path`` is an
        absolute filesystem path to the artifact.
        """
        return self._artifacts


def _save_model_with_class_artifacts_params(
    path,
    python_model,
    artifacts=None,
    conda_env=None,
    code_paths=None,
    mlflow_model=new_model,
):
    """
    :param path: The path to which to save the Python model.
    :param python_model: An instance of a subclass of :class:`~PythonModel`. ``python_model``
                        defines how the model loads artifacts and how it performs inference.
    :param artifacts: A dictionary containing ``<name, artifact_uri>`` entries.
                      Remote artifact URIs
                      are resolved to absolute filesystem paths, producing a dictionary of
                      ``<name, absolute_path>`` entries. ``python_model`` can reference these
                      resolved entries as the ``artifacts`` property of the ``context``
                      attribute. If ``None``, no artifacts are added to the model.
    :param conda_env: Either a dictionary representation of a Conda environment or the
                      path to a Conda environment yaml file. If provided, this decsribes the
                      environment this model should be run in. At minimum, it should specify
                      the dependencies
                      contained in :func:`get_default_conda_env()`. If ``None``, the default
                      :func:`get_default_conda_env()` environment is added to the model.
    :param code_paths: A list of local filesystem paths to Python file dependencies (or
                       directories containing file dependencies). These files are *prepended*
                       to the system path before the model is loaded.
    :param mlflow_model: The model configuration to which to add the ``mlflow.pyfunc`` flavor.
    """
    custom_model_config_kwargs = {
        CONFIG_KEY_CLOUDPICKLE_VERSION: cloudpickle.__version__,
    }
    if isinstance(python_model, PythonModel):
        saved_python_model_subpath = "python_model.pkl"
        with open(os.path.join(path, saved_python_model_subpath), "wb") as out:
            cloudpickle.dump(python_model, out)
        custom_model_config_kwargs[CONFIG_KEY_PYTHON_MODEL] = saved_python_model_subpath
    else:
        raise ClearboxWrapperException(
            "`python_model` must be a subclass of `PythonModel`. Instead, found an"
            " object of type: {python_model_type}".format(
                python_model_type=type(python_model)
            )
        )

    if artifacts:
        saved_artifacts_config = {}
        with TempDir() as tmp_artifacts_dir:
            tmp_artifacts_config = {}
            saved_artifacts_dir_subpath = "artifacts"
            for artifact_name, artifact_uri in artifacts.items():
                tmp_artifact_path = _download_artifact_from_uri(
                    artifact_uri=artifact_uri, output_path=tmp_artifacts_dir.path()
                )
                tmp_artifacts_config[artifact_name] = tmp_artifact_path
                saved_artifact_subpath = posixpath.join(
                    saved_artifacts_dir_subpath,
                    os.path.relpath(
                        path=tmp_artifact_path, start=tmp_artifacts_dir.path()
                    ),
                )
                saved_artifacts_config[artifact_name] = {
                    CONFIG_KEY_ARTIFACT_RELATIVE_PATH: saved_artifact_subpath,
                    CONFIG_KEY_ARTIFACT_URI: artifact_uri,
                }

            shutil.move(
                tmp_artifacts_dir.path(),
                os.path.join(path, saved_artifacts_dir_subpath),
            )
        custom_model_config_kwargs[CONFIG_KEY_ARTIFACTS] = saved_artifacts_config

    conda_env_subpath = "conda.yaml"
    if conda_env is None:
        conda_env = get_default_conda_env()
    elif not isinstance(conda_env, dict):
        with open(conda_env, "r") as f:
            conda_env = yaml.safe_load(f)
    with open(os.path.join(path, conda_env_subpath), "w") as f:
        yaml.safe_dump(conda_env, stream=f, default_flow_style=False)

    saved_code_subpath = None
    if code_paths is not None:
        saved_code_subpath = "code"
        for code_path in code_paths:
            _copy_file_or_tree(src=code_path, dst=path, dst_dir=saved_code_subpath)

    add_pyfunc_flavor_to_model(
        model=mlflow_model,
        loader_module=__name__,
        code=saved_code_subpath,
        env=conda_env_subpath,
        **custom_model_config_kwargs
    )
    mlflow_model.save(os.path.join(path, MLMODEL_FILE_NAME))


def _download_artifact_from_uri(artifact_uri, output_path=None):
    """
    :param artifact_uri: The *absolute* URI of the artifact to download.
    :param output_path: The local filesystem path to which to download the artifact. If
                        unspecified, a local output path will be created.
    """
    parsed_uri = urllib.parse.urlparse(str(artifact_uri))
    prefix = ""
    if parsed_uri.scheme and not parsed_uri.path.startswith("/"):
        # relative path is a special case, urllib does not reconstruct it properly
        prefix = parsed_uri.scheme + ":"
        parsed_uri = parsed_uri._replace(scheme="")

    # For models:/ URIs, it doesn't make sense to initialize a ModelsArtifactRepository with
    # only the model name portion of the URI, then call download_artifacts with the version
    # info.
    if is_models_uri(artifact_uri):
        root_uri = artifact_uri
        artifact_path = ""
    else:
        artifact_path = posixpath.basename(parsed_uri.path)
        parsed_uri = parsed_uri._replace(path=posixpath.dirname(parsed_uri.path))
        root_uri = prefix + urllib.parse.urlunparse(parsed_uri)

    return get_artifact_repository(artifact_uri=root_uri).download_artifacts(
        artifact_path=artifact_path, dst_path=output_path
    )


def is_models_uri(uri):
    return urllib.parse.urlparse(uri).scheme == "models"


def get_artifact_repository(self, artifact_uri):
    """Get an artifact repository from the registry based on the scheme of artifact_uri
    :param store_uri: The store URI. This URI is used to select which artifact repository
                      implementation to instantiate and is passed to the
                      constructor of the implementation.
    :return: An instance of `mlflow.store.ArtifactRepository` that fulfills the artifact URI
             requirements.
    """
    scheme = get_uri_scheme(artifact_uri)
    repository = self._registry.get(scheme)
    if repository is None:
        raise ClearboxWrapperException(
            "Could not find a registered artifact repository for: {}. "
            "Currently registered schemes are: {}".format(
                artifact_uri, list(self._registry.keys())
            )
        )
    return repository(artifact_uri)


def get_uri_scheme(uri_or_path):
    scheme = urllib.parse.urlparse(uri_or_path).scheme
    if any([scheme.lower().startswith(db) for db in DATABASE_ENGINES]):
        return extract_db_type_from_uri(uri_or_path)
    else:
        return scheme


def extract_db_type_from_uri(db_uri):
    """
    Parse the specified DB URI to extract the database type. Confirm the database type is
    supported. If a driver is specified, confirm it passes a plausible regex.
    """
    scheme = urllib.parse.urlparse(db_uri).scheme
    scheme_plus_count = scheme.count("+")

    if scheme_plus_count == 0:
        db_type = scheme
    elif scheme_plus_count == 1:
        db_type, _ = scheme.split("+")
    else:
        error_msg = "Invalid database URI: '%s'. %s" % (db_uri, _INVALID_DB_URI_MSG)
        raise ClearboxWrapperException(error_msg)

    _validate_db_type_string(db_type)

    return db_type


def _validate_db_type_string(db_type):
    """validates db_type parsed from DB URI is supported"""
    if db_type not in DATABASE_ENGINES:
        error_msg = "Invalid database engine: '%s'. '%s'" % (
            db_type,
            _UNSUPPORTED_DB_TYPE_MSG,
        )
        raise ClearboxWrapperException(error_msg)


def _load_pyfunc(model_path):
    pyfunc_config = _get_flavor_configuration(
        model_path=model_path, flavor_name=FLAVOR_NAME
    )

    python_model_cloudpickle_version = pyfunc_config.get(
        CONFIG_KEY_CLOUDPICKLE_VERSION, None
    )
    if python_model_cloudpickle_version is None:
        logger.warning(
            "The version of CloudPickle used to save the model could not be found in"
            " the MLmodel configuration"
        )
    elif python_model_cloudpickle_version != cloudpickle.__version__:
        # CloudPickle does not have a well-defined cross-version compatibility policy. Micro
        # version releases have been known to cause incompatibilities. Therefore, we match
        # on the full library version
        logger.warning(
            "The version of CloudPickle that was used to save the model, `CloudPickle %s`,"
            " differs from the version of CloudPickle that is currently running,"
            " `CloudPickle %s`, and may be incompatible",
            python_model_cloudpickle_version,
            cloudpickle.__version__,
        )

    python_model_subpath = pyfunc_config.get(CONFIG_KEY_PYTHON_MODEL, None)
    if python_model_subpath is None:
        raise ClearboxWrapperException(
            "Python model path was not specified in the model configuration"
        )
    with open(os.path.join(model_path, python_model_subpath), "rb") as f:
        python_model = cloudpickle.load(f)

    artifacts = {}
    for saved_artifact_name, saved_artifact_info in pyfunc_config.get(
        CONFIG_KEY_ARTIFACTS, {}
    ).items():
        artifacts[saved_artifact_name] = os.path.join(
            model_path, saved_artifact_info[CONFIG_KEY_ARTIFACT_RELATIVE_PATH]
        )

    context = PythonModelContext(artifacts=artifacts)
    python_model.load_context(context=context)
    return _PythonModelPyfuncWrapper(python_model=python_model, context=context)


class _PythonModelPyfuncWrapper(object):
    """
    Wrapper class that creates a predict function such that
    predict(model_input: pd.DataFrame) -> model's output as pd.DataFrame (pandas DataFrame)
    """

    def __init__(self, python_model, context):
        """
        :param python_model: An instance of a subclass of :class:`~PythonModel`.
        :param context: A :class:`~PythonModelContext` instance containing artifacts that
                        ``python_model`` may use when performing inference.
        """
        self.python_model = python_model
        self.context = context

    def predict(self, model_input):
        return self.python_model.predict(self.context, model_input)



from typing import Any

import numpy as np
import pandas as pd

from clearbox_wrapper.exceptions import ClearboxWrapperException
from .schema import ColumnSpec, DataType, Schema


class TensorsNotSupportedException(ClearboxWrapperException):
    def __init__(self, msg):
        super().__init__(
            "Multidimensional arrays (aka tensors) are not supported. " "{}".format(msg)
        )


def _infer_schema(data: Any) -> Schema:
    """Infer a schema from data.

    The schema represents data as a sequence of (optionally) named columns with types.

    Parameters
    ----------
    data : Any
        Valid data. It should be one of the following types:
        - pandas.DataFrame or pandas.Series
        - dictionary of { name -> numpy.ndarray}
        - numpy.ndarray
        The data types should be mappable to one of  `clearbox.schema.DataType`.

    Returns
    -------
    Schema
        Schema instance inferred from data.

    Raises
    ------
    TypeError
        If type of data is not valid (pandas.DataFrame or pandas.Series, dictionary of
        { name -> numpy.ndarray}, numpy.ndarray)
    TensorsNotSupportedException
        If data are multidimensional (>2d) arrays (tensors).
    """
    if hasattr(data, "toarray"):
        data = data.toarray()
    elif hasattr(data, "detach"):
        data = data.detach().numpy()
    if isinstance(data, dict):
        res = []
        for col in data.keys():
            ary = data[col]
            if not isinstance(ary, np.ndarray):
                raise TypeError("Data in the dictionary must be of type numpy.ndarray")
            dims = len(ary.shape)
            if dims == 1:
                res.append(ColumnSpec(type=_infer_numpy_array(ary), name=col))
            else:
                raise TensorsNotSupportedException(
                    "Data in the dictionary must be 1-dimensional, "
                    "got shape {}".format(ary.shape)
                )
        schema = Schema(res)
    elif isinstance(data, pd.Series):
        has_nans = data.isna().any()
        series_converted_to_numpy = data.dropna().values if has_nans else data.values
        schema = Schema(
            [
                ColumnSpec(
                    type=_infer_numpy_array(series_converted_to_numpy),
                    has_nans=has_nans,
                )
            ]
        )
    elif isinstance(data, pd.DataFrame):
        columns_spec_list = []
        for col in data.columns:
            has_nans = data[col].isna().any()
            col_converted_to_numpy = (
                data[col].dropna().values if has_nans else data[col].values
            )
            columns_spec_list.append(
                ColumnSpec(
                    type=_infer_numpy_array(col_converted_to_numpy),
                    name=col,
                    has_nans=has_nans,
                )
            )
        schema = Schema(columns_spec_list)
    elif isinstance(data, np.ndarray):
        if len(data.shape) > 2:
            raise TensorsNotSupportedException(
                "Attempting to infer schema from numpy array with "
                "shape {}".format(data.shape)
            )
        if data.dtype == np.object:
            data = pd.DataFrame(data).infer_objects()
            schema = Schema(
                [
                    ColumnSpec(type=_infer_numpy_array(data[col].values))
                    for col in data.columns
                ]
            )
        elif len(data.shape) == 1:
            schema = Schema([ColumnSpec(type=_infer_numpy_dtype(data.dtype))])
        elif len(data.shape) == 2:
            schema = Schema(
                [ColumnSpec(type=_infer_numpy_dtype(data.dtype))] * data.shape[1]
            )
    else:
        raise TypeError(
            "Expected one of (pandas.DataFrame, numpy array, "
            "dictionary of (name -> numpy.ndarray)) "
            "but got '{}'".format(type(data))
        )
    return schema


def _infer_numpy_dtype(dtype: np.dtype) -> DataType:
    """Infer DataType from numpy dtype.

    Parameters
    ----------
    dtype : np.dtype
        Numpy dtype

    Returns
    -------
    DataType
        Inferred DataType.

    Raises
    ------
    TypeError
        If type of `dtype` is not numpy.dtype.
    Exception
        If `dtype.kind`=='O'
    ClearboxWrapperException
        If `dtype` is unsupported.
    """
    if not isinstance(dtype, np.dtype):
        raise TypeError("Expected numpy.dtype, got '{}'.".format(type(dtype)))
    if dtype.kind == "b":
        return DataType.boolean
    elif dtype.kind == "i" or dtype.kind == "u":
        if dtype.itemsize < 4 or (dtype.kind == "i" and dtype.itemsize == 4):
            return DataType.integer
        elif dtype.itemsize < 8 or (dtype.kind == "i" and dtype.itemsize == 8):
            return DataType.long
    elif dtype.kind == "f":
        if dtype.itemsize <= 4:
            return DataType.float
        elif dtype.itemsize <= 8:
            return DataType.double

    elif dtype.kind == "U":
        return DataType.string
    elif dtype.kind == "S":
        return DataType.binary
    elif dtype.kind == "O":
        raise Exception(
            "Can not infer np.object without looking at the values, call "
            "_infer_numpy_array instead."
        )
    raise ClearboxWrapperException(
        "Unsupported numpy data type '{0}', kind '{1}'".format(dtype, dtype.kind)
    )


def _infer_numpy_array(col: np.ndarray) -> DataType:
    """Infer DataType of a numpy array.

    Parameters
    ----------
    col : np.ndarray
        Column representation as a numpy array.

    Returns
    -------
    DataType
        Inferred datatype.

    Raises
    ------
    TypeError
        If `col` is not a numpy array.
    ClearboxWrapperException
        If `col` is not a 1D array.
    """
    if not isinstance(col, np.ndarray):
        raise TypeError("Expected numpy.ndarray, got '{}'.".format(type(col)))
    if len(col.shape) > 1:
        raise ClearboxWrapperException(
            "Expected 1d array, got array with shape {}".format(col.shape)
        )

    class IsInstanceOrNone(object):
        def __init__(self, *args):
            self.classes = args
            self.seen_instances = 0

        def __call__(self, x):
            if x is None:
                return True
            elif any(map(lambda c: isinstance(x, c), self.classes)):
                self.seen_instances += 1
                return True
            else:
                return False

    if col.dtype.kind == "O":
        is_binary_test = IsInstanceOrNone(bytes, bytearray)
        if all(map(is_binary_test, col)) and is_binary_test.seen_instances > 0:
            return DataType.binary
        is_string_test = IsInstanceOrNone(str)
        if all(map(is_string_test, col)) and is_string_test.seen_instances > 0:
            return DataType.string
        # NB: bool is also instance of int => boolean test must precede integer test.
        is_boolean_test = IsInstanceOrNone(bool)
        if all(map(is_boolean_test, col)) and is_boolean_test.seen_instances > 0:
            return DataType.boolean
        is_long_test = IsInstanceOrNone(int)
        if all(map(is_long_test, col)) and is_long_test.seen_instances > 0:
            return DataType.long
        is_double_test = IsInstanceOrNone(float)
        if all(map(is_double_test, col)) and is_double_test.seen_instances > 0:
            return DataType.double
        else:
            raise ClearboxWrapperException(
                "Unable to map 'np.object' type to MLflow DataType. np.object can"
                "be mapped iff all values have identical data type which is one "
                "of (string, (bytes or byterray),  int, float)."
            )
    else:
        return _infer_numpy_dtype(col.dtype)



from abc import ABCMeta, abstractmethod

import cloudpickle

from clearbox_wrapper.utils import _get_default_conda_env


def get_default_conda_env():
    """
    :return: The default Conda environment for MLflow Models produced by calls to
             :func:`save_model() <mlflow.pyfunc.save_model>`
             and :func:`log_model() <mlflow.pyfunc.log_model>` when a user-defined subclass of
             :class:`PythonModel` is provided.
    """
    return _get_default_conda_env(
        additional_conda_deps=None,
        additional_pip_deps=["cloudpickle=={}".format(cloudpickle.__version__)],
        additional_conda_channels=None,
    )


class ClearboxModel(object, metaclass=ABCMeta):
    @abstractmethod
    def prepare_data(self, data):
        pass

    @abstractmethod
    def preprocess_data(self, data):
        pass

    @abstractmethod
    def predict(self, model_input):
        pass

    @abstractmethod
    def predict_proba(self, model_input):
        pass


class _ModelWrapper(object):
    """
    Wrapper class that creates a predict function such that
    predict(model_input: pd.DataFrame) -> model's output as pd.DataFrame (pandas DataFrame)
    """

    def __init__(self, wrapper_model):
        """
        :param python_model: An instance of a subclass of :class:`~PythonModel`.
        :param context: A :class:`~PythonModelContext` instance containing artifacts that
                        ``python_model`` may use when performing inference.
        """
        self.wrapper_model = wrapper_model

    def prepare_data(self, data):
        return self.wrapper_model.prepare_data(data)

    def preprocess_data(self, data):
        return self.wrapper_model.preprocess_data(data)

    def predict(self, model_input):
        return self.wrapper_model.predict(model_input)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['clearbox_wrapper',
 'clearbox_wrapper.data_preparation',
 'clearbox_wrapper.exceptions',
 'clearbox_wrapper.keras',
 'clearbox_wrapper.model',
 'clearbox_wrapper.preprocessing',
 'clearbox_wrapper.pyfunc',
 'clearbox_wrapper.pytorch',
 'clearbox_wrapper.schema',
 'clearbox_wrapper.signature',
 'clearbox_wrapper.sklearn',
 'clearbox_wrapper.utils',
 'clearbox_wrapper.wrapper',
 'clearbox_wrapper.xgboost']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'cloudpickle>=1.6.0,<2.0.0',
 'loguru>=0.5.3,<0.6.0',
 'numpy>=1.16.0,<2.0.0',
 'pandas>=1,<2']

setup_kwargs = {
    'name': 'clearbox-wrapper',
    'version': '0.3.10',
    'description': 'An agnostic wrapper for the most common frameworks of ML models.',
    'long_description': '[![Tests](https://github.com/Clearbox-AI/clearbox-wrapper/workflows/Tests/badge.svg)](https://github.com/Clearbox-AI/clearbox-wrapper/actions?workflow=Tests)\n\n[![PyPI](https://img.shields.io/pypi/v/clearbox-wrapper.svg)](https://pypi.org/project/clearbox-wrapper/)\n\n# Clearbox AI Wrapper\n\nClearbox AI Wrapper is a Python library to package and save a Machine Learning model built with common ML/DL frameworks. It is designed to wrap models trained on strutured data. It includes optional **preprocessing** and **data preparation** functions which can be used to build ready-to-production pipelines.\n\n## Main Features\n\nThe wrapper was born as a fork from [mlflow](https://github.com/mlflow/mlflow) and it\'s based on its [standard format](https://mlflow.org/docs/latest/models.html). It adds the possibility to package, together with the fitted model, preprocessing and data preparation functions in order to create a production-ready pipeline able to receive new data, preprocess them and makes predictions. The resulting wrapped model/pipeline is saved as a zipped folder.\n\nThe library is designed to automatically detect the model framework and its version adding this information to the requirements saved into the final folder. Additional dependencies (e.g. libraries used in preprocessing or data preparation) can also be added as a list parameter if necessary.\n\nThe resulting wrapped folder can be loaded via the Wrapper and the model will be ready to take input through the `predict` or `predict_proba` (if present) method.\n\n**IMPORTANT**: Currently, it is necessary to load the wrapped model with the same Python version with which the model was saved.\n\n## No preprocessing\n\nIn the simplest case, the original dataset has already been preprocessed or it doesn\'t need any preprocessing. It contains only numerical values (ordinal features or one-hot encoded categorical features) and we can easily train a model on it. Then, we only need to save the model and it will be ready to receive new data to make predictions on it.\n\n![](/docs/images/clearbox_ai_wrapper_no_preprocessing.png)\n\nThe following lines show how to wrap and save a simple Scikit-Learn model without preprocessing or data preparation:\n\n```python\nimport clearbox_wrapper as cbw\n\nmodel = DecisionTreeClassifier(max_depth=4, random_state=42)\nmodel.fit(X_train, y_train)\ncbw.save_model(\'wrapped_model_path\', model)\n```\n\n## Preprocessing\n\nTypically, data are preprocessed before being fed into the model. It is almost always necessary to transform (e.g. scaling, binarizing,...) raw data values into a representation that is more suitable for the downstream model. Most kinds of ML models take only numeric data as input, so we must at least encode the non-numeric data, if any.\n\nPreprocessing is usually written and performed separately, before building and training the model. We fit some transformers, transform the whole dataset(s) and train the model on the processed data. If the model goes into production, we need to ship the preprocessing as well. New raw data must be processed on the same way the training dataset was.\n\nWith Clearbox AI Wrapper it\'s possible to wrap and save the preprocessing along with the model so to have a pipeline Processing+Model ready to take raw data, pre-process them and make predictions.\n\n![](/docs/images/clearbox_ai_wrapper_preprocessing.png)\n\nAll the preprocessing code **must be wrapped in a single function** so it can be passed as the `preprocessing` parameter to the `save_model` method. You can use your own custom code for the preprocessing, just remember to wrap all of it in a single function, save it along with the model and add any extra dependencies.\n\n**IMPORTANT**: If the preprocessing includes any kind of fitting on the training dataset (e.g. Scikit Learn transformers), it must be performed outside the final preprocessing function to save. Fit the transformer(s) outside the function and put only the `transform` method inside it. Furthermore, if the entire preprocessing is performed with a single Scikit-Learn transformer, you can directly pass it (fitted) to the `save_model` method.\n\n```python\nfrom sklearn.preprocessing import RobustScaler\nimport xgboost as xgb\n\nimport clearbox_wrapper as cbw\n\n\nx, y = dataset\nx_preprocessor = RobustScaler()\nx_preprocessed = x_preprocessor.fit_transform(x)\n\nmodel = xgb.XGBClassifier(use_label_encoder=False)\nfitted_model = model.fit(x_preprocessed, y)\ncbw.save_model(\'wrapped_model_path\', fitted_model, preprocessing=x_preprocessor)\n```\n\n## Data Preparation (advanced usage)\n\nFor a complex task, a single-step preprocessing could be not enough. Raw data initially collected could be very noisy, contain useless columns or splitted into different dataframes/tables sources. A first data processing is usually performed even before considering any kind of model to feed the data in. The entire dataset is cleaned and the following additional processing and the model are built considering only the cleaned data. But this is not always the case. Sometimes, this situation still applies for data fed in real time to a model in production.\n\nWe believe that a two-step data processing is required to deal with this situation. We refer to the first additional step by the term **Data Preparation**. With Clearbox AI Wrapper it\'s possible to wrap a data preparation step as well, in order to save a final Data Preparation + Preprocessing + Model pipeline ready to takes input.\n\n![](/docs/images/clearbox_ai_wrapper_preprocessing_data_preparation.png)\n\nAll the data preparation code **must be wrapped in a single function** so it can be passed as the `data_preparation` parameter to the `save_model` method. The same considerations wrote above for the preprocessing step still apply for data preparation.\n\n```python\nimport numpy as np\nfrom sklearn.preprocessing import MaxAbsScaler\nfrom tensorflow.keras.layers import Dense\nfrom tensorflow.keras.models import Sequential\n\nimport clearbox_wrapper as cbw\n\ndef preparation(x):\n    data_prepared = np.delete(x, 0, axis=1)\n    return data_prepared\n\nx_preprocessor = RobustScaler()\n\nx, y = dataset\nx_prepared = preparation(x)\nx_preprocessed = x_preprocessor.fit_transform(x_prepared)\n\nmodel = Sequential()\nmodel.add(Dense(8, input_dim=x_preprocessed.shape[1], activation="relu"))\nmodel.add(Dense(3, activation="softmax"))\n\nmodel.compile(\n    optimizer="adam", loss="sparse_categorical_crossentropy", metrics=["accuracy"]\n)\nmodel.fit(x_preprocessed, y)\n\ncbw.save_model(\n    \'wrapped_model_path\',\n    model,\n    preprocessing=x_preprocessor,\n    data_preparation=preparation\n)\n```\n\n### Data Preparation vs. Preprocessing\n\nIt is not always clear which are the differences between preprocessing and data preparation. It\'s not easy to understand where data preparation ends and preprocessing begins. There are no conditions that apply in any case, but in general you should build the data preparation step working only with the dataset, without considering the model your data will be fed into. Any kind of operation is allowed, but often preparing the raw data includes removing or normalizing some columns, replacing values, add a column based on other column values,... After this step, no matter what kind of transformation the data have been through, they should still be readable and understandable by a human user.\n\nThe preprocessing step, on the contrary, should be considered closely tied with the downstream ML model and adapted to its particular "needs". Typically processed data by this second step are only numeric and non necessarily understandable by a human.\n\n## Supported ML frameworks\n\n- Scikit-Learn\n- XGBoost\n- Keras\n- Pytorch\n\n## Installation\n\nInstall the latest relased version on the [Python Package Index (PyPI)](https://pypi.org/project/clearbox-wrapper/) with\n\n```shell\npip install clearbox-wrapper\n```\n## Examples\n\nThe following Jupyter notebooks provide examples of simle and complex cases:\n\n- [Scikit Learn Decision Tree on Iris Dataset](https://github.com/Clearbox-AI/clearbox-wrapper/blob/master/examples/1_iris_sklearn/1_Clearbox_Wrapper_Iris_Scikit.ipynb) (No preprocessing, No data preparation)\n- [XGBoost Model on Lending Club Loans Dataset](https://github.com/Clearbox-AI/clearbox-wrapper/blob/master/examples/2_loans_preprocessing_xgboost/2_Clearbox_Wrapper_Loans_Xgboost.ipynb) (Preprocessing, No data preparation)\n- [Pytorch Network on Boston Housing Dataset](https://github.com/Clearbox-AI/clearbox-wrapper/blob/master/examples/3_boston_preprocessing_pytorch/3_Clearbox_Wrapper_Boston_Pytorch.ipynb) (Preprocessing, No data preparation)\n- [Keras Network on UCI Adult Dataset](https://github.com/Clearbox-AI/clearbox-wrapper/blob/master/examples/4_adult_data_cleaning_preprocessing_keras/4_Clearbox_Wrapper_Adult_Keras.ipynb) (Preprocessing and data preparation)\n- [Pytorch Network on Diabetes Hospital Readmissions](https://github.com/Clearbox-AI/clearbox-wrapper/blob/master/examples/5_hospital_preprocessing_pytorch/5_Clearbox_Wrapper_Hospital_Pytorch.ipynb) (Preprocessing and data preparation)\n\n## License\n\n[Apache License 2.0](https://github.com/Clearbox-AI/clearbox-wrapper/blob/master/LICENSE)\n',
    'author': 'Clearbox AI',
    'author_email': 'info@clearbox.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://clearbox.ai',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



from typing import Any, Dict, Optional, Union

import numpy as np
import pandas as pd

from clearbox_wrapper.schema import _infer_schema, Schema

InferableDataset = Union[pd.DataFrame, np.ndarray, Dict[str, np.ndarray]]


class Signature(object):
    """Description of a Model, Preprocessing or Data Preparation inputs and outpus.

    Attributes
    ----------
    inputs : clearbox_wrapper.schema.Schema
        Inputs schema as a sequence of (optionally) named columns with types.
    outputs : Optional[clearbox_wrapper.schema.Schema]
        Outputs schema as a sequence of (optionally) named columns with types.

    """

    def __init__(self, inputs: Schema, outputs: Optional[Schema] = None) -> "Signature":
        """Create a new Signature instance given inputs and (optionally) outputs schemas.

        Parameters
        ----------
        inputs : clearbox_wrapper.schema.Schema
            Inputs schema as a sequence of (optionally) named columns with types.
        outputs : Optional[clearbox_wrapper.schema.Schema]
            Outputs schema as a sequence of (optionally) named columns with types,
            by default None.


        Raises
        ------
        TypeError
            If `inputs` is not type Schema or if `outputs` is not type Schema or None.
        """
        if not isinstance(inputs, Schema):
            raise TypeError("Inputs must be type Schema, got '{}'".format(type(inputs)))
        if outputs is not None and not isinstance(outputs, Schema):
            raise TypeError(
                "Outputs must be either None or clearbox_wrapper.schema.Schema, "
                "got '{}'".format(type(inputs))
            )
        self.inputs = inputs
        self.outputs = outputs

    def to_dict(self) -> Dict[str, Any]:
        """Generate dictionary representation of the signature.

        Returns
        -------
        Dict[str, Any]
            Signature dictionary {"inputs": inputs schema as JSON,
            "outputs": outputs schema as JSON}
        """
        return {
            "inputs": self.inputs.to_json(),
            "outputs": self.outputs.to_json() if self.outputs is not None else None,
        }

    @classmethod
    def from_dict(cls, signature_dict: Dict[str, Any]) -> "Signature":
        """Create a Signature instance from a dictionary representation.

        Parameters
        ----------
        signature_dict: Dict[str, Any]
            Signature dictionary {"inputs": inputs schema as JSON,
            "outputs": outputs schema as JSON}

        Returns
        -------
        Signature
            A Signature instance.
        """
        inputs = Schema.from_json(signature_dict["inputs"])
        if "outputs" in signature_dict and signature_dict["outputs"] is not None:
            outputs = Schema.from_json(signature_dict["outputs"])
            return cls(inputs, outputs)
        else:
            return cls(inputs)

    def __eq__(self, other: "Signature") -> bool:
        """Check if two Signature instances (self and other) are equal.

        Parameters
        ----------
        other : Signature
            A Signature instance

        Returns
        -------
        bool
            True if the two signatures are equal, False otherwise.
        """
        return (
            isinstance(other, Signature)
            and self.inputs == other.inputs
            and self.outputs == other.outputs
        )

    def __repr__(self) -> str:
        """Generate string representation.

        Returns
        -------
        str
            Signature string representation.
        """
        return (
            "inputs: \n"
            "  {}\n"
            "outputs: \n"
            "  {}\n".format(repr(self.inputs), repr(self.outputs))
        )


def infer_signature(input_data: Any, output_data: InferableDataset = None) -> Signature:
    """Infer a Signature from input data and (optionally) output data.

    The signature represents inputs and outputs scheme as  a sequence
    of (optionally) named columns with types.

    Parameters
    ----------
    input_data : Any
        Valid input data. E.g. (a subset of) the training dataset. It should be
        one of the following types:
        - pandas.DataFrame
        - dictionary of { name -> numpy.ndarray}
        - numpy.ndarray
        The element types should be mappable to one of `clearbox.schema.DataType`.
    output_data : InferableDataset, optional
        Valid output data. E.g. Preprocessed data or model predictions for
        (a subset of) the training dataset. It should be one of the following types:
        - pandas.DataFrame
        - numpy.ndarray
        The element types should be mappable to one of `clearbox.schema.DataType`.
        By default None.

    Returns
    -------
    Signature
        Inferred Signature
    """
    inputs = _infer_schema(input_data)
    outputs = _infer_schema(output_data) if output_data is not None else None
    return Signature(inputs, outputs)
